//////////////////////////////////////////////////////////////////////////
//                                                                      //
// expf.cpp -                                                           // 
//                                                                      //
//                              //
//        .                          //
//                                                                      //
// Copyright (c) 2001 RC Module Inc.                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
//   $Revision:: 1     $      $Date:: 2/14/06 8:07p    $                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "math.h"
#include "auxfd.h"

//      exp

const int _TLN2 = 0x2C5C85FE;

//  Ln(1+1/2)*c, Ln(1+1/4)*c, Ln(1+1/8)*c, ...
//  c = 040000000
static int _TLN[29] ={0x19F323EC,	0xE47FBE3,	0x789C1DB,	0x3E14618,
		0x1F829B0,	0xFE0545,	0x7F80A9,	0x3FE015,
		0x1FF802,	0xFFE00,	0x7FF80,	0x3FFE0,
		0x1FFF8,	0xFFFE,	0x7FFF,	0x3FFF,	0x1FFF,	0xFFF,	0x7FF,
		0x3FF,	0x1FF,	0xFF,	0x7F,	0x3F,	0x1F,	0xF,	0x7,
		0x3,	0x1	};

#ifdef __cplusplus
extern "C" {
#endif

// expf --
//
//    
//
//      --.
//      CORDIC,    .
float expf(float f){
	int Mnt;                            //   
    int iExp, Sgn;                      //     
    int NewExp;                         //   
    int Spc;
	int tmp; 
	unsigned int rez = 0x40000000;      //    1
	unsigned int Arg;                   //     

    //   float
	DestructFloat( f, Mnt, iExp, Sgn, Spc );

	if ( iExp > 6 ){			//    
		//  
		if ( iExp ==128 ){
			if ( Mnt!=0 )					//  4 NAN
				return f;
			else if ( Sgn ==0 )	//  1 +inf
				return f;
			else if ( Sgn ==1 )	//  2 -inf
				return 0;
		}

		if ( Sgn==0 ){
			if (!_FPIEEE_Flags)
				_FPIEEE_Signal( _FPIEEE_Overflow );
			_FPIEEE_Flags |= _FPIEEE_Overflow;
			return ConstructFloat( 0, 0x80, 0, 0);
		}
		else{
			if (!_FPIEEE_Flags)
				_FPIEEE_Signal( _FPIEEE_Underflow );
			_FPIEEE_Flags |= _FPIEEE_Underflow;
			return ConstructFloat( 0, -127, 0, 0);		
		}
	}

			
	if ( iExp < -24 ){			//    
		//  
		if ( iExp == -127 && Mnt!=0 ){				//  3	
			return 1;
		}

		return ConstructFloat( 0, 0, 0, 0);		
	}

    //       ,
    //    -   
	int i=-8;

    //     () ,  floor

    //      ,  - Arg * 2^i
	if (iExp>=0){
		Arg = Mnt << (iExp);
	}
	else if ( iExp >= -2 ){
		Arg = Mnt << 4;
		Arg = Arg >> (-iExp);
		i=-4;
	}
	else {
		Arg = Mnt << 8;
		Arg = Arg >> (-iExp);
		i=0;
	}

	NewExp = 0;
	tmp = Arg;      //        
    
    //   ,   

    //     .     
    //  2^i * Ln(2),      2^i. ( -i)
	while (i<0){  //  	
		i++;
		tmp -= _TLN2;
		if (tmp>0){
			Arg = tmp;
			NewExp += 1 << (-i);
		}
		else //    - 
			tmp = Arg;
		tmp <<=1;
		Arg = tmp;
	}
	tmp >>=1;   //    ,    
	Arg = tmp;
    //  ,      
	if (NewExp>127){
		if (Sgn==1){
			if(NewExp>156){
            //       , ret=0
    			_FPIEEE_Signal( _FPIEEE_Underflow );
				return ConstructFloat( 0, -127, 0, 0);
			}
		}
		else{
            //   , ret=+inf
			_FPIEEE_Signal( _FPIEEE_Overflow );
			return ConstructFloat( 0, 0x80, 0, 0);
		}
	}
    //   ( exp(x) = a * 2^b )    exp(-x) = ( Ln(2)-a ) * 2^(-b-1)
	if (Sgn==1){
		Arg = _TLN2 - Arg;
		tmp = Arg;
		NewExp = -(++NewExp);
	}

    //      CORDIC    
	while (i!=30){  //  
		tmp -= _TLN[i++];   //    ,
                            //    ,  _TLN[]
                            //  ,    
		if (tmp>0){
			rez += rez >> (i);  //   -   (1 + (0.5)^i)
			Arg = tmp;
		}
		else //    - 
			tmp = Arg;
	}

	if (NewExp<-126){	//  -  
		rez >>= (-127 - NewExp)+1;
		NewExp = -127;
	}

	rez >>=6;	//    , ...
	rez = (rez>>1) + (rez & 1); //   ,  .

	if (rez == 0x01000000){	
		rez = 0;
		NewExp++;
	}

    //   float
    return ConstructFloat( rez, NewExp, 0, 0);	
}

#ifdef __cplusplus
}
#endif
