#ifndef	_MALLOC_INTERNAL
#define _MALLOC_INTERNAL
#include <malloc.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#endif

#define min(A, B) ((A) < (B) ? (A) : (B))


static __malloc_size_t  __size_by_ptr(__ptr_t ptr,_heap_control_block *cb)
{
    int type;
    __malloc_size_t block;

    assert(ptr!=NULL);
    block = BLOCK (ptr,cb);
    type=cb->_heapinfo[block].busy.type;

    //switch (type)
    //{
    //    case 0: return cb->_heapinfo[block].busy.info.size*BLOCKSIZE;
    //    default: return (1 << type);
    //}
	return type ? (1 << type) : cb->_heapinfo[block].busy.info.size*BLOCKSIZE;
}

static __ptr_t __realloc_internal(__ptr_t ptr, __malloc_size_t size,_heap_control_block *cb)
{
        int s;
        __ptr_t p;
        p=__malloc_internal(size,cb);
        if (!p) return NULL;
        s=__size_by_ptr(ptr,cb);
        memmove(p,ptr,min(s,size));
        __free_internal(ptr,cb);
        return p;
}

extern "C" __ptr_t realloc0(__ptr_t ptr, __malloc_size_t size)
{
        __ptr_t p;
        assert(!ptr || __IN_HEAP0(ptr));
        p=__realloc_internal(ptr,size,&_hcb0);
        assert(!p || __IN_HEAP0(p));
        return p;
}

extern "C" __ptr_t realloc1(__ptr_t ptr, __malloc_size_t size)
{
        __ptr_t p;
        assert(!ptr || __IN_HEAP1(ptr));
        p=__realloc_internal(ptr,size,&_hcb1);
        assert(!p || __IN_HEAP1(p));
        return p;
}

extern "C" __ptr_t realloc2(__ptr_t ptr, __malloc_size_t size)
{
        __ptr_t p;
        assert(!ptr || __IN_HEAP2(ptr));
        p=__realloc_internal(ptr,size,&_hcb2);
        assert(!p || __IN_HEAP2(p));
        return p;
}

extern "C" __ptr_t realloc3(__ptr_t ptr, __malloc_size_t size)
{
        __ptr_t p;
        assert(!ptr || __IN_HEAP3(ptr));
        p=__realloc_internal(ptr,size,&_hcb3);
        assert(!p || __IN_HEAP3(p));
        return p;
}

inline __ptr_t alloc_somewhere( __ptr_t ptr, __malloc_size_t size, _heap_control_block *cb )
{
	__ptr_t p = malloc(size);
	if(p){
        memmove( p, ptr, __size_by_ptr( ptr, cb) );		
	    __free_internal( ptr, cb );
	}
	return p;
}

extern "C" __ptr_t realloc(__ptr_t ptr, __malloc_size_t size)
{
  __ptr_t p;

  if (ptr==NULL) return malloc(size);
  if (__IN_HEAP0(ptr))
  {
     /* trying to reallocate localy */
     p=realloc0(ptr,size);

     if (p) return p;
     /* trying to reallocate to different heap */
     return alloc_somewhere( ptr, size, &_hcb0 );
  }
  else if (__IN_HEAP1(ptr))
  {
     /* trying to reallocate localy */
     p=realloc1(ptr,size);

     if (p) return p;
     /* trying to reallocate to different heap */
     return alloc_somewhere( ptr, size, &_hcb1 );
  }
  else if (__IN_HEAP2(ptr))
  {
     /* trying to reallocate localy */
     p=realloc2(ptr,size);

     if (p) return p;
     /* trying to reallocate to different heap */
     return alloc_somewhere( ptr, size, &_hcb2 );
  }
  else if (__IN_HEAP3(ptr))
  {
     /* trying to reallocate localy */
     p=realloc3(ptr,size);

     if (p) return p;
     /* trying to reallocate to different heap */
     return alloc_somewhere( ptr, size, &_hcb3 );
  }
  return NULL;
}
